package io.scanbot.sdk.flutter

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.net.Uri
import android.util.Log
import com.google.gson.FieldNamingPolicy
import com.google.gson.Gson
import com.google.gson.GsonBuilder
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.scanbot.sdk.barcode.ScanbotBarcodeDetector
import io.scanbot.sdk.barcode.entity.BarcodeDocumentFormat
import io.scanbot.sdk.barcode.entity.BarcodeFormat
import io.scanbot.sdk.barcode.entity.BarcodeFormattedData
import io.scanbot.sdk.barcode.entity.BarcodeItem
import io.scanbot.sdk.barcode.entity.BarcodeScannerAdditionalConfig
import io.scanbot.sdk.barcode.entity.BarcodeScanningResult
import io.scanbot.sdk.barcode.entity.FormattedBarcodeDataMapper
import io.scanbot.sdk.barcode.entity.MSIPlesseyChecksumAlgorithm
import io.scanbot.sdk.flutter.utils.getBitmapFromContentUri
import io.scanbot.sdk.flutter.utils.getBitmapFromFileUri
import io.scanbot.sdk.flutter.ISdkSubplugin
import io.scanbot.sdk.persistence.BarcodeFileStorage
import io.scanbot.sdk.persistence.fileio.FileIOProcessor
import io.scanbot.sdk.ui.view.barcode.InternalBarcodeScannerActivity
import io.scanbot.sdk.ui.view.barcode.batch.configuration.BatchBarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerAdditionalConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.base.RtuConstants
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.withContext
import java.io.File
import java.util.ArrayList
import java.util.EnumSet
import kotlin.coroutines.resume
import kotlin.coroutines.resumeWithException
import kotlin.coroutines.suspendCoroutine

class SharedLogicPlugin(
    val activity: Activity,
    var barcodeCallbackChannel: MethodChannel?,
    coroutineResultHandler: CoroutineResultHandler,
    val gson: Gson,
    val getBarcodeDetector: () -> ScanbotBarcodeDetector,
    val getBarcodeStorage: () -> BarcodeFileStorage,
    val getFileIOProcessor: () -> FileIOProcessor,
    val getBarcodeScannerIntent: (Activity, BarcodeScannerConfiguration) -> Intent,
    val getBatchBarcodeScannerIntent: (Activity, BatchBarcodeScannerConfiguration) -> Intent
) : ISdkSubplugin,
    ICoroutineActivityResultHandler by coroutineResultHandler {
    override suspend fun onMethodCall(call: MethodCall, result: MethodChannel.Result): Boolean {
        when (call.method) {
            "cleanupBarcodeStorage" -> {
                cleanupBarcodeStorage(activity, call, result)
            }

            "startBarcodeScanner" -> {
                proceedWithBarcodeScanner(activity, call, result)
            }

            "startBatchBarcodeScanner" -> {
                proceedWithBatchBarcodeScanner(activity, call, result)
            }

            "detectBarcodesOnImages" -> {
                detectBarcodesOnImages(activity, call, result)
            }

            "getDecryptedDataFromFile" -> {
                getDecryptedDataFromFile(activity, call, result)
            }

            else -> return false
        }
        return true
    }

    init {
        batchBarcodeCallbackChannel = barcodeCallbackChannel
    }

    companion object {
        @JvmStatic
        var batchBarcodeCallbackChannel: MethodChannel? = null

        @JvmStatic
        private val gson = GsonBuilder()
            .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
            .registerTypeAdapter(kotlin.ByteArray::class.java, ByteArrayToBase64TypeAdapter())
            .create()

        @JvmStatic
        private suspend fun getFormattedData(barcodeItem: BarcodeItem) =
            suspendCoroutine<BarcodeFormattedData> {
                val json = gson.toJson(barcodeItem)
                batchBarcodeCallbackChannel?.invokeMethod(
                    "getFormattedData",
                    json.toString(),
                    object : MethodChannel.Result {
                        override fun success(result: Any?) {
                            val json = (result as BarcodeFormattedJson)
                            val data = BarcodeFormattedData(
                                (json["title"] as? String)
                                    ?: barcodeItem.barcodeFormat.name,
                                (json["subtitle"] as? String) ?: barcodeItem.text ?: ""
                            );
                            it.resume(data)
                        }

                        override fun error(
                            errorCode: String,
                            errorMessage: String?,
                            errorDetails: Any?
                        ) {
                            it.resumeWithException(IllegalStateException(errorMessage))
                        }

                        override fun notImplemented() {
                            it.resumeWithException(IllegalStateException("Not yet implemented"))
                        }
                    })
            }
    }

    private suspend fun cleanupBarcodeStorage(
        activity: Activity,
        call: MethodCall,
        methodResult: MethodChannel.Result
    ) {
        withContext(Dispatchers.Main) {
            try {

                getBarcodeStorage().cleanupBarcodeImagesDirectory();
                methodResult.success("");
            } catch (e: Exception) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(
                    "Sdk exception",
                    e.localizedMessage,
                    null
                )
            }
        }
    }

    private suspend fun proceedWithBarcodeScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: MethodChannel.Result
    ) {
        val barcodeCameraConfiguration = try {
            readBarcodeScannerConfiguration(call)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) {
                methodResult.error(
                    e::class.java.simpleName,
                    e.message,
                    null
                )
            }
            return
        }
        val intent =
            getBarcodeScannerIntent(activity, barcodeCameraConfiguration)
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<BarcodeScanningResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                val imagePath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = it.toJson(gson, imagePath, previewPath)
                    withContext(Dispatchers.Main) {
                        methodResult.success(
                            gson.toJson(
                                result
                            )
                        )
                    }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }


    class FlutterInnerBarcodeFormatter : FormattedBarcodeDataMapper {
        override fun decodeFormattedData(barcodeItem: BarcodeItem): BarcodeFormattedData {
            return try {
                runBlocking {
                    withContext(Dispatchers.Main) {
                        getFormattedData(
                            barcodeItem
                        )
                    }
                }
            } catch (e: Throwable) {
                BarcodeFormattedData(barcodeItem.barcodeFormat.name, barcodeItem.text)
            }

        }
    }

    private suspend fun proceedWithBatchBarcodeScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: MethodChannel.Result
    ) {
        val barcodeCameraConfiguration = try {
            readBatchBarcodeScannerConfiguration(call)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) {
                methodResult.error(
                    e::class.java.simpleName,
                    e.message,
                    null
                )
            }
            return
        }

        val arguments = call.arguments as BarcodeScannerConfigurationJson
        val isNeedToSetCallback = (arguments["shouldUseBarcodeFormatter"] as? Boolean) ?: false
        if (isNeedToSetCallback) {
            barcodeCameraConfiguration.setFormattedBarcodeDataMapper(FlutterInnerBarcodeFormatter())
        }

        val intent = getBatchBarcodeScannerIntent(activity, barcodeCameraConfiguration)

        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<BarcodeScanningResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                val imagePath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = it.toJson(gson, imagePath, previewPath)
                    withContext(Dispatchers.Main) {
                        methodResult.success(
                            gson.toJson(
                                result
                            )
                        )
                    }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }


    /// common method used for detecting the barcodes
    private suspend fun detectBarcodesOnImages(
        activity: Activity,
        call: MethodCall,
        methodResult: MethodChannel.Result
    ) {
        //too add high sensitivity mode to and remove rotation
        withContext(Dispatchers.Default) {
            try {
                val barcodeDetector = setUpBarcodeDetector(call);
                val uris = call.argument<ArrayList<String>>("uris")
                if (uris == null || uris.isEmpty()) {
                    methodResult.error("IllegalArgumentException", "Need to setup file uri =", null)
                } else {
                    var barcodeResults = ArrayList<Map<String, Any>>();
                    for (uriStringItem in uris) {
                        val uri = Uri.parse(uriStringItem)
                        val file = File(uri.path)
                        val bitmap = when (uri.scheme) {
                            "content" -> {
                                getBitmapFromContentUri(activity.contentResolver, uri)
                            }

                            "file" -> {
                                getBitmapFromFileUri(file)
                            }

                            else -> {
                                if (file.exists()) {
                                    getBitmapFromFileUri(file)
                                } else {
                                    barcodeResults.add(
                                        getBarcodeDetectionError(
                                            uriStringItem,
                                            "Unknown file path format."
                                        )
                                    );
                                    return@withContext;
                                }
                            }
                        }

                        bitmap?.let {
                            val barcodeData = barcodeDetector.detectFromBitmap(bitmap, 0)
                            barcodeData?.let { barcodeDataResult ->
                                var successData = mapOf<String, Any>(
                                    "barcodeImageURI" to uriStringItem,
                                    "barcodeItems" to barcodeData.barcodeItems
                                );

                                barcodeResults.add(successData);
                            }
                            if (barcodeData == null || barcodeData.barcodeItems?.count() == 0) {
                                barcodeResults.add(
                                    getBarcodeDetectionError(
                                        uriStringItem,
                                        "No barcodes found."
                                    )
                                );
                            }
                        }

                        if (bitmap == null) {
                            barcodeResults.add(
                                getBarcodeDetectionError(
                                    uriStringItem,
                                    "bitmap not found."
                                )
                            );
                        }
                    }

                    var result = mapOf("barcodeResults" to barcodeResults);
                    methodResult.success(gson.toJson(result));
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error("IllegalStateException", e.message, null)
                }
            }
        }
    }


    /// get the error message payload
    private fun getBarcodeDetectionError(uri: String, errorMessage: String): Map<String, Any> {
        return mapOf<String, Any>(
            "barcodeImageURI" to uri,
            "error" to errorMessage,
            "barcodeItems" to ArrayList<String>()
        );
    }

    /// set up the barcode detector object and returns the same.
    private fun setUpBarcodeDetector(call: MethodCall): ScanbotBarcodeDetector {
        var barcodeDetector = getBarcodeDetector()
        val arguments = call.arguments as Json
        if (call.hasArgument("barcodeFormats")) {
            val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
            val acceptedBarcodeFormats =
                barcodeFormats.map { BarcodeFormat.valueOf(it) }
            barcodeDetector.modifyConfig {
                this.setBarcodeFormats(acceptedBarcodeFormats)
            }
        }

        if (call.hasArgument("acceptedDocumentFormats")) {
            val acceptedDocumentFormats =
                arguments["acceptedDocumentFormats"] as ArrayList<String>
            val documentFormats =
                acceptedDocumentFormats.map { BarcodeDocumentFormat.valueOf(it) }
            barcodeDetector.modifyConfig {
                this.setAcceptedDocumentFormats(documentFormats)
            }
        }

        if (call.hasArgument("additionalParameters")) {
            val arguments = call.arguments as Json
            val barcodeAdditionalParameters =
                arguments["additionalParameters"] as Json
            barcodeDetector.modifyConfig {
                this.setAdditionalConfig(
                    gson.fromJson(
                        barcodeAdditionalParameters.toString(),
                        BarcodeScannerAdditionalConfig::class.java
                    )
                )
            }
        }

        return barcodeDetector;
    }

    private suspend fun getDecryptedDataFromFile(
        context: Context,
        call: MethodCall,
        methodResult: MethodChannel.Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val imageUriString = call.argument<String?>("fileUri")
                if (imageUriString.isNullOrEmpty()) {
                    withContext(Dispatchers.Main) {
                        methodResult.error(
                            "SdkParamException",
                            "imageFileUri is null or empty",
                            null
                        )
                    }
                    return@withContext
                }

                val file = File(Uri.parse(imageUriString).path!!)
                if (!file.exists()) {
                    withContext(Dispatchers.Main) {
                        methodResult.error(
                            "FileAccessException",
                            "\"File not found",
                            null
                        )
                    }
                    return@withContext
                }
                val fileIOProcessor = getFileIOProcessor()
                val decodedFile = fileIOProcessor.read(file)
                // In case image file encryption is enabled, the corresponding FileIOProcessor
                // implementation will decrypt image data under the hood:
                withContext(Dispatchers.Main) { methodResult.success(decodedFile) }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "RuntimeException",
                        "error decoding image from file",
                        null
                    )
                }
            }
        }
    }

    private suspend fun dispatchCanceled(methodResult: MethodChannel.Result) {
        withContext(Dispatchers.Main) {
            methodResult.success("CANCELED")
        }
    }


}