package io.scanbot.sdk.flutter.camera

import android.content.Context
import android.content.MutableContextWrapper
import android.view.View
import android.widget.FrameLayout
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.platform.PlatformView
import io.scanbot.sdk.flutter.ByteArrayToBase64TypeAdapter
import io.scanbot.sdk.flutter.Json
import io.scanbot.sdk.barcode.ui.BarcodeScannerView
import io.scanbot.sdk.barcode_scanner.ScanbotBarcodeScannerSDK
import io.scanbot.sdk.flutter.camera.detectorWrappers.BarcodeDetector
import io.scanbot.sdk.flutter.camera.detectorWrappers.IDetectorWrapper
import io.scanbot.sdk.flutter.checkParam
import io.scanbot.sdk.flutter.utils.LifecycleOwnerWrapper
import io.scanbot.sdk.flutter.utils.getLicenseStatusJson
import io.scanbot.sdk.ui.camera.CustomCameraLifecycleOwner
import io.scanbot.sdk.util.view.unwrapLifecycleOwner
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch
import kotlin.coroutines.CoroutineContext


class ScanbotCameraPlatformView internal constructor(
    private val context: Context,
    private val messenger: BinaryMessenger,
    private val id: Int,
    private val creationParams: Json?
) : PlatformView,
    MethodChannel.MethodCallHandler, CoroutineScope {

    private var rootView: FrameLayout = FrameLayout(context)
    private val scanbotSDK = ScanbotBarcodeScannerSDK(context)
    private val customViewDisplayOwner: CustomCameraLifecycleOwner = CustomCameraLifecycleOwner()
    private var cameraChannel = MethodChannel(messenger, "scanbot_sdk_camera_$id")
    private val detectorChannel = MethodChannel(messenger, "scanbot_sdk_camera.detector_$id")
    private val finderChannel = MethodChannel(
        messenger,
        "scanbot_sdk_camera.finder_${creationParams?.get("layout_orientation") ?: ""}"
    )
    private var detector: IDetectorWrapper? = null
    private val returnContext = Dispatchers.Main

    val gson = GsonBuilder()
        .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
        .registerTypeAdapter(ByteArray::class.java, ByteArrayToBase64TypeAdapter())
        .create()

    private var job: Job = Job()
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.Default + job

    override fun getView(): View {
        return rootView
    }

    init {
        initCamera()
    }

    private fun initCamera() {
        cameraChannel.setMethodCallHandler(this)
        creationParams?.apply {
            checkParam<Json>("scannerConfiguration")?.let {
                attachDetector(it)
            }
        }
        customViewDisplayOwner.startCamera()
    }

    override fun onMethodCall(methodCall: MethodCall, result: MethodChannel.Result) {
        when (methodCall.method) {
            "flashEnabled" -> {
                val useFlash = methodCall.argument<Boolean>("enabled") ?: false
                detector?.flashEnabled(useFlash)
                result.success("success")
            }

            "toggleContinuousFocus" -> {
                detector?.toggleContinuousFocus()
                result.success("success")
            }

            "pause" -> {
                customViewDisplayOwner.stopCamera()
                result.success("success")
            }

            "resume" -> {
                customViewDisplayOwner.startCamera()
                result.success("success")
            }

            "dispose" -> {
                customViewDisplayOwner.stopCamera()
                dispose()
                result.success("success")
            }

            "isFlashAvailable" -> {
                result.success(detector?.isFlashAvailable() ?: false)
            }

            else -> result.notImplemented()
        }
    }

    private fun attachDetector(arguments: Json) {
        clearPrevDetectorState()
        if (checkLicenseInvalid()) return
        detector = BarcodeScannerView(context).let {
            LifecycleOwnerWrapper(
                context,
                it,
                (context as MutableContextWrapper).baseContext.unwrapLifecycleOwner()!!
            ).let { wrapper ->
                rootView.addView(
                    wrapper,
                    FrameLayout.LayoutParams(
                        FrameLayout.LayoutParams.MATCH_PARENT,
                        FrameLayout.LayoutParams.MATCH_PARENT
                    )
                )
                BarcodeDetector(
                    wrapper, this,
                    cameraChannel,
                    detectorChannel,
                    finderChannel,
                    scanbotSDK.createBarcodeDetector(),
                    scanbotSDK.createBarcodeFileStorage(),
                    scanbotSDK.fileIOProcessor(),
                    customViewDisplayOwner,
                    creationParams ?: HashMap(),
                    gson,
                    { scanbotSDK.licenseInfo }
                )
            }

        }


        detector?.attach(arguments)
    }

    private fun checkLicenseInvalid(): Boolean {
        if (!scanbotSDK.licenseInfo.isValid) {
            launch(returnContext) {
                val json = scanbotSDK.licenseInfo.getLicenseStatusJson()
                detectorChannel.invokeMethod(
                    "onLicenseError",
                    json.toString()
                )
            }
            return true
        }
        return false
    }

    private fun clearPrevDetectorState() {
        rootView.removeAllViews()
        detector?.detach()
        detector = null
    }


    override fun dispose() {
        clearPrevDetectorState()
        job.cancel()
        customViewDisplayOwner.destroyCamera()
    }

}

