import 'dart:async';

import 'package:barcode_scanner/shared_calls.dart';
import 'package:barcode_scanner/scanbot_sdk_models.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:logging/logging.dart';

import 'barcode_scanning_data.dart';
import 'license.dart';

export 'classical_components/hint.dart';
export 'classical_components/shutter.dart';
export 'classical_components/classical_camera.dart';
export 'classical_components/camera_configuration.dart';
export 'classical_components/barcode_camera.dart';
export 'json/common.dart';
export 'json/common_platform.dart';
export 'barcode_scanning_data.dart';
export 'scanbot_sdk_models.dart';
export 'sdk_utils.dart';
export 'license.dart';
export 'scanbot_encryption_handler.dart';

class ScanbotBarcodeSdk {
  static const MethodChannel _channel = MethodChannel('barcode_scanner');

  /// Initializes the Scanbot SDK with given [config]s.
  /// The Scanbot SDK must be initialized before using any other API methods.
  static Future initScanbotSdk(ScanbotSdkConfig config) async {
    try {
      LicenseRegistry.addLicense(licenses);
      var json = config.toJson();
      await _channel.invokeMethod('initScanbotSdk', json);
    } catch (e) {
      Logger.root.severe(e);
      rethrow;
    }
  }

  /// Returns the current license status of the Scanbot SDK.
  static Future<SdkLicenseStatus> getLicenseStatus() {
    return SharedCalls.getLicenseStatusInternal(_channel);
  }

  /// Clear storage folder with barcode snapped images
  static Future cleanupBarcodeStorage() {
    return SharedCalls.cleanupBarcodeStorageInternal(_channel);
  }

  /// Starts the Ready-to-Use Barcode Scanner UI as full-screen modal UI dialog.
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes (QR codes, etc).
  static Future<BarcodeScanningResult> startBarcodeScanner(
      BarcodeScannerConfiguration config) async {
    return SharedCalls.startBarcodeScannerInternal(_channel, config);
  }

  /// Starts the Ready-to-Use Batch Barcode Scanner UI as full-screen modal UI dialog.
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes (QR codes, etc).
  static Future<BarcodeScanningResult> startBatchBarcodeScanner(
      BatchBarcodeScannerConfiguration config) async {
    return SharedCalls.startBatchBarcodeScanner(_channel, config);
  }

  /// Detect barcodes from image file uri.
  ///
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes.
  ///
  /// [barcodeFormats] barcode type(s) that should be recognized.
  ///
  /// [acceptedDocumentFormats] barcode document type(s) that should be recognized.
  static Future<BarcodeScanningResult> detectBarcodesOnImage(Uri file,
      {List<BarcodeFormat> barcodeFormats = const <BarcodeFormat>[],
      BarcodeAdditionalParameters? additionalParameters,
      List<BarcodeDocumentFormat> acceptedDocumentFormats =
          const <BarcodeDocumentFormat>[]}) async {
    return await SharedCalls.detectBarcodesFromImage(_channel, file,
        barcodeFormats: barcodeFormats,
        additionalParameters: additionalParameters,
        acceptedDocumentFormats: acceptedDocumentFormats);
  }

  /// Detect barcodes from the given images list
  static Future<MultiImageBarcodeScanningResult> detectBarcodesOnImages(
      List<Uri> imageFileUris,
      {List<BarcodeFormat> barcodeFormats = const <BarcodeFormat>[],
      BarcodeAdditionalParameters? additionalParameters,
      List<BarcodeDocumentFormat> acceptedDocumentFormats =
          const <BarcodeDocumentFormat>[]}) async {
    return await SharedCalls.detectBarcodesFromImages(_channel, imageFileUris,
        barcodeFormats: barcodeFormats,
        additionalParameters: additionalParameters,
        acceptedDocumentFormats: acceptedDocumentFormats);
  }
}

typedef Handler = Future Function(MethodCall call);
