import 'dart:convert';
import 'dart:io';

import 'package:flutter/services.dart';
import 'package:logging/logging.dart';

import 'barcode_scanning_data.dart';
import 'meta.dart';
import 'scanbot_sdk_models.dart';

class SharedCalls {
  static const MethodChannel _batchBarcodeFormatterCallbackChannel =
      MethodChannel('batch_barcode_data_mapper');

  static Handler? _handler;

  static void _initBatchFormatterHandler(
      MethodChannel sessionCallbackChannel, BarcodeDataFormatter formatter) {
    _handler = (MethodCall call) async {
      if (call.method == 'getFormattedData') {
        var dataItem =
            await formatter(BarcodeItem.fromJson(jsonDecode(call.arguments)));
        return dataItem.toJson();
      }
    };
  }

  /// Returns the current license status of the Scanbot SDK.
  static Future<SdkLicenseStatus> getLicenseStatusInternal(
      MethodChannel channel) async {
    try {
      var licenseStatus = await channel.invokeMethod('getLicenseStatus', {});
      var json = jsonDecode(licenseStatus);
      return SdkLicenseStatus.fromJson(json);
    } catch (e) {
      Logger.root.severe(e);
      rethrow;
    }
  }

  /// Detect barcodes from image file uri.
  ///
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes.
  ///
  /// [barcodeFormats] barcode type(s) that should be recognized.
  ///
  /// [acceptedDocumentFormats] barcode document type(s) that should be recognized.
  static Future<BarcodeScanningResult> detectBarcodesFromImage(
    MethodChannel channel,
    Uri file, {
    List<BarcodeFormat> barcodeFormats = const <BarcodeFormat>[],
    BarcodeAdditionalParameters? additionalParameters,
    List<BarcodeDocumentFormat> acceptedDocumentFormats =
        const <BarcodeDocumentFormat>[],
  }) async {
    var barcodeResult = BarcodeScanningResult.error(
        "Barcode detection failed"); // default error message initialisation
    try {
      List<Uri> list = [file];
      var result = await detectBarcodesFromImages(
        channel,
        list,
        barcodeFormats: barcodeFormats,
        additionalParameters: additionalParameters,
        acceptedDocumentFormats: acceptedDocumentFormats,
      );
      if (result.operationResult == OperationResult.SUCCESS) {
        if (result.barcodeResults.isNotEmpty) {
          barcodeResult = result.barcodeResults.first;
        }
      } else {
        barcodeResult.message = result.message;
      }
      return barcodeResult;
    } catch (e) {
      Logger.root.severe(e);
      barcodeResult.message = e.toString();
      return barcodeResult;
    }
  }

  /// Detect barcodes from the given images list
  static Future<MultiImageBarcodeScanningResult> detectBarcodesFromImages(
      MethodChannel channel, List<Uri> imageFileUris,
      {List<BarcodeFormat> barcodeFormats = const <BarcodeFormat>[],
      BarcodeAdditionalParameters? additionalParameters,
      List<BarcodeDocumentFormat> acceptedDocumentFormats =
          const <BarcodeDocumentFormat>[]}) async {
    try {
      var detectFromMultipleImageParams = DetectBarcodeFromImageParams(
          barcodeFormats: barcodeFormats,
          acceptedDocumentFormats: acceptedDocumentFormats,
          additionalParameters: additionalParameters,
          uris: imageFileUris);
      var barcodeScanningResult = await channel.invokeMethod(
          'detectBarcodesOnImages', detectFromMultipleImageParams.toJson());
      var result = MultiImageBarcodeScanningResult.fromJson(
          jsonDecode(barcodeScanningResult));
      return result;
    } catch (e) {
      Logger.root.severe(e);
      return MultiImageBarcodeScanningResult.error("parsing exception");
    }
  }

  /// Starts the Ready-to-Use Barcode Scanner UI as full-screen modal UI dialog.
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes (QR codes, etc).
  static Future<BarcodeScanningResult> startBarcodeScannerInternal(
      MethodChannel channel, BarcodeScannerConfiguration config) async {
    try {
      Logger.root.log(Level.INFO, "BARCODE init formats");
      if (config.barcodeFormats == null) {
        config.barcodeFormats = PredefinedBarcodes.allBarcodeTypes();
      } else if (config.barcodeFormats!.isEmpty) {
        config.barcodeFormats = PredefinedBarcodes.allBarcodeTypes();
      }
      Logger.root.log(Level.INFO, "BARCODE config to JSON");

      var updatedParameters = config.toJson();
      const defaultJsonName = "BarcodeScreen";
      Logger.root.log(Level.INFO, "BARCODE update default JSON");
      var configuration = await applyParametersOnDefaultJson(
          defaultJsonName, updatedParameters);
      Logger.root.log(Level.INFO, "BARCODE run scanner");
      var result = await invoke(channel, 'startBarcodeScanner', configuration);
      Logger.root.log(Level.INFO, "BARCODE scanner canseled");

      if (wasCanceled(result)) {
        return BarcodeScanningResult.canceled();
      }
      return BarcodeScanningResult.fromJson(jsonDecode(result));
    } catch (e) {
      Logger.root.severe(e);
      return BarcodeScanningResult.error(e.toString());
    }
  }

  /// Starts the Ready-to-Use Batch Barcode Scanner UI as full-screen modal UI dialog.
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes (QR codes, etc).
  static Future<BarcodeScanningResult> startBatchBarcodeScanner(
      MethodChannel channel, BatchBarcodeScannerConfiguration config) async {
    try {
      if (config.barcodeFormats == null) {
        config.barcodeFormats = PredefinedBarcodes.allBarcodeTypes();
      } else if (config.barcodeFormats!.isEmpty) {
        config.barcodeFormats = PredefinedBarcodes.allBarcodeTypes();
      }
      var updatedParameters = config.toJson();

      if (config.barcodeFormatter != null) {
        _initBatchFormatterHandler(
            _batchBarcodeFormatterCallbackChannel, config.barcodeFormatter!);
        _batchBarcodeFormatterCallbackChannel.setMethodCallHandler(_handler);
        updatedParameters['shouldUseBarcodeFormatter'] = true;
      } else {
        _handler = null;
      }

      const defaultJsonName = "BatchBarcodeScreen";
      var configuration = await applyParametersOnDefaultJson(
          defaultJsonName, updatedParameters);
      var barcodeScanningResult =
          await channel.invokeMethod('startBatchBarcodeScanner', configuration);
      _handler = null;
      if (wasCanceled(barcodeScanningResult)) {
        return BarcodeScanningResult.canceled();
      }
      return BarcodeScanningResult.fromJson(jsonDecode(barcodeScanningResult));
    } catch (e) {
      Logger.root.severe(e);
      return BarcodeScanningResult.error(e.toString());
    }
  }

  /// This method takes the default json file and updates the parameters with the given parameters from customer configuration
  /// The approach is needed to always have the full json come to the native platform
  static Future<Map<String, dynamic>> applyParametersOnDefaultJson(
      String defaultJsonName, Map<String, dynamic> updatedParameters) async {
    var defaultConfiguration = jsonDecode(await rootBundle.loadString(
            'packages/' +
                scanbot_library_name +
                '/assets/jsonExamples/$defaultJsonName.json'))
        as Map<String, dynamic>;

    if (Platform.isAndroid) {
      checkColors(defaultConfiguration);
    }

    updateParameter(updatedParameters, defaultConfiguration);
    return defaultConfiguration;
  }

  static void updateParameter(Map<String, dynamic> updatedParameters,
      Map<String, dynamic> defaultConfiguration) {
    updatedParameters.forEach((key, value) {
      if (value is Map) {
        var map = value as Map<String, dynamic>;
        updateParameter(map, defaultConfiguration[key]);
      } else {
        defaultConfiguration[key] = value;
      }
    });
  }

  /// take the color all strings with alpha and swap the last two characters to the start position RGBA-> ARGB
  static void checkColors(Map<String, dynamic> defaultConfiguration) {
    defaultConfiguration.forEach((key, value) {
      if (value is Map) {
        checkColors(defaultConfiguration[key]);
      } else {
        defaultConfiguration[key] = swapColorString(value);
      }
    });
  }

  static dynamic swapColorString(dynamic value) {
    if (value is String) {
      final hexColorRegExp = RegExp(r'^#([A-Fa-f0-9]{8})$');
      if (hexColorRegExp.hasMatch(value)) {
        String lastTwoChars = value.substring(value.length - 2);
        String withoutLastTwoChars = value.substring(0, value.length - 2);
        return '#' + lastTwoChars + withoutLastTwoChars.substring(1);
      }
      return value;
    }
    return value;
  }

  /// Clear storage folder with barcode snapped images
  static Future cleanupBarcodeStorageInternal(MethodChannel channel) async {
    try {
      await channel.invokeMethod('cleanupBarcodeStorage', {});
      return;
    } catch (e) {
      Logger.root.severe(e);
      rethrow;
    }
  }

  static Future<String> invoke(
      MethodChannel _channel, String method, Map<String, dynamic> json) async {
    return await _channel.invokeMethod(method, json);
  }

  static bool wasCanceled(String result) {
    return result == 'CANCELED';
  }
}

typedef Handler = Future Function(MethodCall call);
